from enum import Enum

class Network(Enum):
    """ Enum for the different virtual costs the user can choose, they map to the Java equivalent class name for easy mapping
    """
    MACROSCOPIC = "org.planit.network.physical.macroscopic.MacroscopicNetwork"
    PHYSICAL = "org.planit.network.physical.PhysicalNetwork"
    VIRTUAL = "org.planit.network.virtual.VirtualNetwork"
    
class OutputFormatter(Enum):
    """ Enum for the different output formatters the user can choose, they map to the Java equivalent class name for easy mapping
        Only the output formatters available in the PLANitIO project have been defined here
    """
    
    MEMORY = "org.planit.output.formatter.MemoryOutputFormatter"
    PLANIT_IO = "org.planit.io.output.formatter.PlanItOutputFormatter"
    
class OutputProperty(Enum):    
    """ Enum for the different output properties the user can configure in the output files 
        Equivalent of Java enumeration org.planit.output.property.OutputProperty
    """
    
    # alphabetical order by priority type (ID, INPUT, OUTPUT
    DESTINATION_ZONE_ID = "DESTINATION_ZONE_ID"                     # ID
    DESTINATION_ZONE_EXTERNAL_ID = "DESTINATION_ZONE_EXTERNAL_ID"   # ID
    DOWNSTREAM_NODE_ID = "DOWNSTREAM_NODE_ID"                       # ID
    DOWNSTREAM_NODE_EXTERNAL_ID = "DOWNSTREAM_NODE_EXTERNAL_ID"     # ID
    ITERATION_INDEX = "ITERATION_INDEX"                             # ID
    LINK_SEGMENT_ID = "LINK_SEGMENT_ID"                             # ID
    LINK_SEGMENT_EXTERNAL_ID = "LINK_SEGMENT_EXTERNAL_ID"           # ID
    MODE_ID = "MODE_ID"                                             # ID
    MODE_EXTERNAL_ID = "MODE_EXTERNAL_ID"                           # ID
    ORIGIN_ZONE_ID = "ORIGIN_ZONE_ID"                               # ID
    ORIGIN_ZONE_EXTERNAL_ID = "ORIGIN_ZONE_EXTERNAL_ID"             # ID
    PATH_ID = "PATH_ID"                                                #ID
    RUN_ID = "RUN_ID"                                               # ID
    TIME_PERIOD_ID = "TIME_PERIOD_ID"                               # ID
    TIME_PERIOD_EXTERNAL_ID = "TIME_PERIOD_EXTERNAL_ID"             # ID
    UPSTREAM_NODE_ID = "UPSTREAM_NODE_ID"                           # ID
    UPSTREAM_NODE_EXTERNAL_ID = "UPSTREAM_NODE_EXTERNAL_ID"         # ID
    CAPACITY_PER_LANE = "CAPACITY_PER_LANE"                         # INPUT
    DOWNSTREAM_NODE_LOCATION = "DOWNSTREAM_NODE_LOCATION"           # INPUT
    LENGTH = "LENGTH"                                               # INPUT
    LINK_TYPE = "LINK_TYPE"                                         # INPUT    
    MAXIMUM_DENSITY = "MAXIMUM_DENSITY"       # INPUT
    MAXIMUM_SPEED = "MAXIMUM_SPEED"                                 # INPUT
    NUMBER_OF_LANES = "NUMBER_OF_LANES"                             # INPUT
    UPSTREAM_NODE_LOCATION = "UPSTREAM_NODE_LOCATION"               # INPUT        
    CALCULATED_SPEED = "CALCULATED_SPEED"                           # OUTPUT
    COST_TIMES_FLOW = "COST_TIMES_FLOW"                             # OUTPUT
    DENSITY = "DENSITY"                                             # OUTPUT        
    FLOW = "FLOW"                                                   # OUTPUT
    LINK_COST = "LINK_COST"                                         # OUTPUT
    OD_COST = "OD_COST"                                             # OUTPUT
    TOTAL_COST_TO_END_NODE = "TOTAL_COST_TO_END_NODE"               # OUTPUT
    PATH_STRING = "PATH_STRING"                               # OUTPUT
    VC_RATIO = "VC_RATIO"                                           # OUTPUT
        
    
    def java_class_name(self) -> str:
        return "org.planit.output.property.OutputProperty"     
    
class OutputType(Enum):
    """ Enum for the different output types the user can choose to activate, 
         Equivalent of Java enumeration org.planit.output.OutputType
    """
    LINK = "LINK"
    GENERAL = "GENERAL"
    SIMULATION = "SIMULATION"
    OD = "OD"
    PATH = "PATH"
    
    def java_class_name(self) -> str:
        return "org.planit.output.enums.OutputType"   
    
class PhysicalCost(Enum):
    """ Enum for the different physical costs the user can choose, they map to the Java equivalent class name for easy mapping
    """
    BPR = "org.planit.cost.physical.BPRLinkTravelTimeCost"

class RouteIdType(Enum):

    LINK_SEGMENT_EXTERNAL_ID = "LINK_SEGMENT_EXTERNAL_ID"
    LINK_SEGMENT_ID = "LINK_SEGMENT_ID"
    NODE_EXTERNAL_ID = "NODE_EXTERNAL_ID"
    NODE_ID = "NODE_ID"

    def java_class_name(self) -> str:
        return "org.planit.output.enums.RouteIdType"     
   
class Smoothing(Enum):
    """ Enum for the different smoothing options the user can choose, they map to the Java equivalent class name for easy mapping
    """
    MSA = "org.planit.sdinteraction.smoothing.MSASmoothing"

class TrafficAssignment(Enum):
    """ Enum for the different assignment the user can choose, they map to the Java equivalent class name for easy mapping
    """
    TRADITIONAL_STATIC = "org.planit.trafficassignment.TraditionalStaticAssignment"
    ETLM = "org.planit.ltm.trafficassignment.ETLM"

class VirtualCost(Enum):
    """ Enum for the different virtual costs the user can choose, they map to the Java equivalent class name for easy mapping
    """
    FIXED = "org.planit.cost.virtual.FixedConnectoidTravelTimeCost"
    SPEED = "org.planit.cost.virtual.SpeedConnectiodTravelTimeCost"
    
class ODSkimSubOutputType(Enum):
    
    NONE = "NONE"
    COST = "COST"

    def java_class_name(self) -> str:
        return "org.planit.output.enums.ODSkimSubOutputType"     



import os
import pandas as pd
from planit import TrafficAssignment
from planit import OutputType
from planit import OutputProperty
from planit import RouteIdType
from planit import OutputFormatter
from planit import ODSkimSubOutputType
from builtins import staticmethod

class PlanItHelper:
    
    @staticmethod
    def run_test5(plan_it, max_iterations, epsilon, description, output_type_configuration_option, project_path=None, deactivate_file_output=False):
        """Top-level method which runs unit test case 5
        :param plan_it PLANit object (with any required initial costs already defined
        :param max_iterations the maximum number of iterations for the current unit test
        :param epsilon the convergence epsilon for the current unit test
        :param description the name to be used to identify input and output files
        :param output_type_configuration_option used to specify which properties to remove from link output type configuration
        :param project_path directory of XML input file (if omitted, defaults to None which will make methods use the current directory)
        :param deactivate_file_output if True, deactivate the file output formatter and store results in memory only
        """        
        plan_it.set(TrafficAssignment.TRADITIONAL_STATIC)         
        plan_it.assignment.physical_cost.set_default_parameters(0.8, 4.5, 2, 1)

        # TODO : Add a unit test which testsX plan_it.assigment.physical_cost.set_default_parameters()
        # testRouteChoiceCompareWithOmniTRANS5() is a  good one for this
        #plan_it.assignment.set(VirtualCost.FIXED)
        #plan_it.assignment.set(Smoothing.MSA)
        plan_it.assignment.output_configuration.set_persist_only_final_Iteration(True)
        plan_it.assignment.activate_output(OutputType.LINK)
        plan_it.assignment.link_configuration.remove(OutputProperty.TIME_PERIOD_EXTERNAL_ID)
        plan_it.assignment.link_configuration.remove(OutputProperty.TIME_PERIOD_ID)
  
        if output_type_configuration_option == 1:
            plan_it.assignment.link_configuration.remove(OutputProperty.MAXIMUM_SPEED)
        elif output_type_configuration_option == 2:
            plan_it.assignment.link_configuration.remove(OutputProperty.DOWNSTREAM_NODE_EXTERNAL_ID)
            plan_it.assignment.link_configuration.remove(OutputProperty.UPSTREAM_NODE_EXTERNAL_ID)
  
        plan_it.assignment.activate_output(OutputType.OD)
        plan_it.assignment.od_configuration.deactivate(ODSkimSubOutputType.NONE)
        plan_it.assignment.od_configuration.remove(OutputProperty.TIME_PERIOD_EXTERNAL_ID)
        plan_it.assignment.od_configuration.remove(OutputProperty.RUN_ID)
        plan_it.assignment.activate_output(OutputType.PATH)
        plan_it.assignment.path_configuration.set_path_id_type(RouteIdType.NODE_EXTERNAL_ID)
        plan_it.assignment.gap_function.stop_criterion.set_max_iterations(max_iterations)
        plan_it.assignment.gap_function.stop_criterion.set_epsilon(epsilon)
         
        plan_it.activate(OutputFormatter.MEMORY)
        if deactivate_file_output:
            plan_it.deactivate(OutputFormatter.PLANIT_IO)
        else:
            plan_it.output.set_xml_name_root(description)                
            plan_it.output.set_csv_name_root(description)     
            if (project_path is not None):  
                plan_it.output.set_output_directory(project_path)
        plan_it.run()
        return plan_it
   
    @staticmethod
    def run_mode_test(plan_it, max_iterations, epsilon, description, output_type_configuration_option, project_path=None, deactivate_file_output=False):
        """Top-level method which runs unit mode test
        :param plan_it PLANit object (with any required initial costs already defined
        :param max_iterations the maximum number of iterations for the current unit test
        :param epsilon the convergence epsilon for the current unit test
        :param description the name to be used to identify input and output files
        :param output_type_configuration_option used to specify which properties to remove from link output type configuration
        :param project_path directory of XML input file (if omitted, defaults to None which will make methods use the current directory)
        :param deactivate_file_output if True, deactivate the file output formatter and store results in memory only
        """        
        plan_it.set(TrafficAssignment.TRADITIONAL_STATIC)         
        # TODO : Add a unit test which testsX plan_it.assigment.physical_cost.set_default_parameters()
        # testRouteChoiceCompareWithOmniTRANS5() is a  good one for this
        plan_it.assignment.output_configuration.set_persist_only_final_Iteration(True)
        plan_it.assignment.activate_output(OutputType.LINK)
        plan_it.assignment.link_configuration.remove(OutputProperty.TIME_PERIOD_EXTERNAL_ID)
        plan_it.assignment.link_configuration.remove(OutputProperty.TIME_PERIOD_ID)
  
        if output_type_configuration_option == 1:
            plan_it.assignment.link_configuration.remove(OutputProperty.MAXIMUM_SPEED)
        elif output_type_configuration_option == 2:
            plan_it.assignment.link_configuration.remove(OutputProperty.DOWNSTREAM_NODE_EXTERNAL_ID)
            plan_it.assignment.link_configuration.remove(OutputProperty.UPSTREAM_NODE_EXTERNAL_ID)
  
        plan_it.assignment.activate_output(OutputType.OD)
        plan_it.assignment.od_configuration.deactivate(ODSkimSubOutputType.NONE)
        plan_it.assignment.od_configuration.remove(OutputProperty.TIME_PERIOD_EXTERNAL_ID)
        plan_it.assignment.od_configuration.remove(OutputProperty.RUN_ID)
        plan_it.assignment.activate_output(OutputType.PATH)
        plan_it.assignment.path_configuration.set_path_id_type(RouteIdType.NODE_EXTERNAL_ID)
        plan_it.assignment.gap_function.stop_criterion.set_max_iterations(max_iterations)
        plan_it.assignment.gap_function.stop_criterion.set_epsilon(epsilon)
         
        plan_it.activate(OutputFormatter.MEMORY)
        if deactivate_file_output:
            plan_it.deactivate(OutputFormatter.PLANIT_IO)
        else:
            plan_it.output.set_xml_name_root(description)                
            plan_it.output.set_csv_name_root(description)     
            if (project_path is not None):  
                plan_it.output.set_output_directory(project_path)
                
        plan_it.assignment.physical_cost.set_default_parameters(0.8, 4.5, 1, 1)
        link_segment_external_id = 3
        mode_external_id = 1 
        plan_it.assignment.physical_cost.set_parameters(1.0, 5.0, mode_external_id, link_segment_external_id)       
                
        plan_it.run()
        return plan_it
 
    @staticmethod
    def run_test(plan_it, max_iterations, epsilon, description, output_type_configuration_option, project_path=None, deactivate_file_output=False):
        """Top-level method which runs unit tests
        :param plan_it PLANit object (with any required initial costs already defined
        :param max_iterations the maximum number of iterations for the current unit test
        :param epsilon the convergence epsilon for the current unit test
        :param description the name to be used to identify input and output files
        :param output_type_configuration_option used to specify which properties to remove from link output type configuration
        :param project_path directory of XML input file (if omitted, defaults to None which will make methods use the current directory)
        :param deactivate_file_output if True, deactivate the file output formatter and store results in memory only
        """        
        plan_it.set(TrafficAssignment.TRADITIONAL_STATIC)         
        # TODO : Add a unit test which testsX plan_it.assigment.physical_cost.set_default_parameters()
        # testRouteChoiceCompareWithOmniTRANS5() is a  good one for this
        plan_it.assignment.output_configuration.set_persist_only_final_Iteration(True)
        plan_it.assignment.activate_output(OutputType.LINK)
        plan_it.assignment.link_configuration.remove(OutputProperty.TIME_PERIOD_EXTERNAL_ID)
        plan_it.assignment.link_configuration.remove(OutputProperty.TIME_PERIOD_ID)
  
        if output_type_configuration_option == 1:
            plan_it.assignment.link_configuration.remove(OutputProperty.MAXIMUM_SPEED)
        elif output_type_configuration_option == 2:
            plan_it.assignment.link_configuration.remove(OutputProperty.DOWNSTREAM_NODE_EXTERNAL_ID)
            plan_it.assignment.link_configuration.remove(OutputProperty.UPSTREAM_NODE_EXTERNAL_ID)
  
        plan_it.assignment.activate_output(OutputType.OD)
        plan_it.assignment.od_configuration.deactivate(ODSkimSubOutputType.NONE)
        plan_it.assignment.od_configuration.remove(OutputProperty.TIME_PERIOD_EXTERNAL_ID)
        plan_it.assignment.od_configuration.remove(OutputProperty.RUN_ID)
        plan_it.assignment.activate_output(OutputType.PATH)
        plan_it.assignment.path_configuration.set_path_id_type(RouteIdType.NODE_EXTERNAL_ID)
        plan_it.assignment.gap_function.stop_criterion.set_max_iterations(max_iterations)
        plan_it.assignment.gap_function.stop_criterion.set_epsilon(epsilon)
         
        plan_it.activate(OutputFormatter.MEMORY)
        if deactivate_file_output:
            plan_it.deactivate(OutputFormatter.PLANIT_IO)
        else:
            plan_it.output.set_xml_name_root(description)                
            plan_it.output.set_csv_name_root(description)     
            if (project_path is not None):  
                plan_it.output.set_output_directory(project_path)
        plan_it.run()
        return plan_it
  
    @staticmethod
    def run_test_with_zero_flow_outputs(plan_it, max_iterations, epsilon, description, output_type_configuration_option, project_path=None, deactivate_file_output=False):
        """Top-level method which runs unit tests with zero flow outputs included
        :param plan_it PLANit object (with any required initial costs already defined
        :param max_iterations the maximum number of iterations for the current unit test
        :param epsilon the convergence epsilon for the current unit test
        :param description the name to be used to identify input and output files
        :param output_type_configuration_option used to specify which properties to remove from link output type configuration
        :param project_path directory of XML input file (if omitted, defaults to None which will make methods use the current directory)
        :param deactivate_file_output if True, deactivate the file output formatter and store results in memory only
        """        
        plan_it.set(TrafficAssignment.TRADITIONAL_STATIC)         
        # TODO : Add a unit test which testsX plan_it.assigment.physical_cost.set_default_parameters()
        # testRouteChoiceCompareWithOmniTRANS5() is a  good one for this
        plan_it.assignment.output_configuration.set_persist_only_final_Iteration(True)
        plan_it.assignment.output_configuration.set_persist_zero_flow(True)
        
        plan_it.assignment.activate_output(OutputType.LINK)
        plan_it.assignment.link_configuration.remove(OutputProperty.TIME_PERIOD_EXTERNAL_ID)
        plan_it.assignment.link_configuration.remove(OutputProperty.TIME_PERIOD_ID)
  
        if output_type_configuration_option == 1:
            plan_it.assignment.link_configuration.remove(OutputProperty.MAXIMUM_SPEED)
        elif output_type_configuration_option == 2:
            plan_it.assignment.link_configuration.remove(OutputProperty.DOWNSTREAM_NODE_EXTERNAL_ID)
            plan_it.assignment.link_configuration.remove(OutputProperty.UPSTREAM_NODE_EXTERNAL_ID)
  
        plan_it.assignment.activate_output(OutputType.OD)
        plan_it.assignment.od_configuration.deactivate(ODSkimSubOutputType.NONE)
        plan_it.assignment.od_configuration.remove(OutputProperty.TIME_PERIOD_EXTERNAL_ID)
        plan_it.assignment.od_configuration.remove(OutputProperty.RUN_ID)
        plan_it.assignment.activate_output(OutputType.PATH)
        plan_it.assignment.path_configuration.set_path_id_type(RouteIdType.NODE_EXTERNAL_ID)
        plan_it.assignment.gap_function.stop_criterion.set_max_iterations(max_iterations)
        plan_it.assignment.gap_function.stop_criterion.set_epsilon(epsilon)
         
        plan_it.activate(OutputFormatter.MEMORY)
        if deactivate_file_output:
            plan_it.deactivate(OutputFormatter.PLANIT_IO)
        else:
            plan_it.output.set_xml_name_root(description)                
            plan_it.output.set_csv_name_root(description)     
            if (project_path is not None):  
                plan_it.output.set_output_directory(project_path)
        plan_it.run()
        return plan_it

    @staticmethod
    def run_test_without_activating_outputs(plan_it, max_iterations, epsilon, description, output_type_configuration_option, project_path=None, deactivate_file_output=False):
        """Top-level method which runs unit tests without activating output configurations (these should come on automatically)
        :param plan_it PLANit object (with any initial costs already defined)
        :param max_iterations the maximum number of iterations for the current unit test
        :param epsilon the convergence epsilon for the current unit test
        :param description the name to be used to identify input and output files
        :param output_type_configuration_option used to specify which properties to remove from link output type configuration
        :param project_path directory of XML input file (if omitted, defaults to None which will make methods use the current directory)
        :param deactivate_file_output if True, deactivate the file output formatter and store results in memory only
        """
        
        plan_it.set(TrafficAssignment.TRADITIONAL_STATIC)
         
        # TODO : Add a unit test which testsX plan_it.assigment.physical_cost.set_default_parameters()
        # testRouteChoiceCompareWithOmniTRANS5() is a  good one for this
        plan_it.assignment.output_configuration.set_persist_only_final_Iteration(True)
        plan_it.assignment.link_configuration.remove(OutputProperty.TIME_PERIOD_EXTERNAL_ID)
        plan_it.assignment.link_configuration.remove(OutputProperty.TIME_PERIOD_ID)
  
        if output_type_configuration_option == 1:
            plan_it.assignment.link_configuration.remove(OutputProperty.MAXIMUM_SPEED)
        elif output_type_configuration_option == 2:
            plan_it.assignment.link_configuration.remove(OutputProperty.DOWNSTREAM_NODE_EXTERNAL_ID)
            plan_it.assignment.link_configuration.remove(OutputProperty.UPSTREAM_NODE_EXTERNAL_ID)
  
        #Note that OutputType.OD is deactivated, check later that no OD output file is created
        plan_it.assignment.activate_output(OutputType.OD)
        plan_it.assignment.deactivate_output(OutputType.OD)
        
        plan_it.assignment.path_configuration.set_path_id_type(RouteIdType.NODE_EXTERNAL_ID)
        plan_it.assignment.gap_function.stop_criterion.set_max_iterations(max_iterations)
        plan_it.assignment.gap_function.stop_criterion.set_epsilon(epsilon)
         
        plan_it.activate(OutputFormatter.MEMORY)
        if deactivate_file_output:
            plan_it.deactivate(OutputFormatter.PLANIT_IO)
        else:
            plan_it.output.set_xml_name_root(description)                
            plan_it.output.set_csv_name_root(description)     
            if (project_path is not None):  
                plan_it.output.set_output_directory(project_path)
 
        plan_it.run()
        plan_it.assignment.activate_output(OutputType.OD)
        return plan_it

    
    @staticmethod
    def delete_file(output_type : OutputType, description, file_name, project_path=None):
        """Delete an output file
        :param output_type type of the output file (link, origin-destination or path)
        :param description root name of the output file
        :param project_path directory of the output file
        """
        if project_path == None:
            project_path = os.getcwd()
        full_file_name = PlanItHelper.create_full_file_name(output_type, project_path, description, file_name)
        os.remove(full_file_name)
        
    @staticmethod
    def create_full_file_name(output_type : OutputType, project_path, description, file_name):
        """Create the long name of the output file (containing results created by the current test run)
        :param output_type type of the output file (link, origin-destination or path)
        :param description root name of the output file
        :param project_path directory of the output file
        :param file_name name of the file of standard results
        :return the name of the file containing the test results
        """
        type_name = None
        if output_type.value == "LINK":
            type_name = 'Link'
        elif output_type.value == "OD":
            type_name = 'Origin-Destination'
        else:
            type_name = "Path"
            
        full_file_name = project_path 
        full_file_name += "\\" 
        full_file_name +=  type_name
        full_file_name +=  "_" 
        full_file_name +=  "RunId_0_"
        full_file_name +=  description 
        full_file_name +=  "_" 
        full_file_name += file_name
        return full_file_name
    
    @staticmethod
    def create_short_file_name(output_type : OutputType, project_path, file_name):
        """Create the short name of the output file (containing standard results)
        :param output_type type of the output file (link, origin-destination or path)
        :param project_path directory of the output file
        :param file_name name of the file of standard results
        :return the name of the file containing the standard results
        """
        if output_type.value == "LINK":
            type_name = 'Link'
        elif output_type.value == "OD":
            type_name = 'Origin-Destination'
        else:
            type_name = "Path"
            
        short_file_name = project_path 
        short_file_name += "\\" 
        short_file_name +=  type_name
        short_file_name +=  "_" 
        short_file_name += file_name
        return short_file_name
    
    @staticmethod
    def compare_csv_files(csv_file_location1, csv_file_location2):
        """Compare the contents of two CSV files, returning true if they are equal, false otherwise
        :param csv_file_location1 first CSV file to be compared
        :param csv_file_location2 second CSV file to be compared
        :return true if the files have equal contents, false otherwise
        """
        df1 = pd.read_csv(csv_file_location1)
        df2 = pd.read_csv(csv_file_location2)
        return df1.equals(df2)
    
    @staticmethod
    def compare_csv_files_and_clean_up(output_type : OutputType, description, file_name, project_path=None):
        """Compare the file of test results with the file of standard results, and delete the test results file if they are equal
        :param output_type type of the output file (link, origin-destination or path)
        :param description root name of the output file
        :param project_path directory of the output file
        :param file_name name of the file of standard results
        :return true if the files have equal contents and the results file has been deleted, false otherwise
        """
        if project_path == None:
            project_path = os.getcwd()
        full_file_name = PlanItHelper.create_full_file_name(output_type, project_path, description, file_name)
        short_file_name = PlanItHelper.create_short_file_name(output_type, project_path, file_name)
        comparison_result = PlanItHelper.compare_csv_files(short_file_name, full_file_name)
        if (comparison_result):
            os.remove(full_file_name)
        return comparison_result


import os
import subprocess
import traceback
from py4j.java_gateway import JavaGateway
from planit import GatewayUtils
from planit import GatewayState
from planit import GatewayConfig
from planit import BaseWrapper 
from planit import MacroscopicNetworkWrapper 
from planit import DemandsWrapper
from planit import AssignmentWrapper
from planit import ZoningWrapper
from planit import PlanItOutputFormatterWrapper
from planit import MemoryOutputFormatterWrapper
from planit import TrafficAssignment
from planit import OutputFormatter
from planit import InitialCostWrapper
from planit import TimePeriodWrapper
from planit import InitialCost
from builtins import isinstance

class PLANit:
            
    def __init__(self, project_path=None, debug_info=False, standalone=True):
        """Constructor of PLANit python wrapper which acts as an interface to the underlying PLANit Java code
        :param project_path the path location of the XML input file(s) to be used by PLANitIO
        :param standalone when true this PLANit instance bootstraps a java gateway and closes it upon completion of the scripts when false <to be implemented>
        """  
        # explicitly set uninitialized member variables to None
        self._assignment_instance = None
        self._input_builder_instance = None
        self._network_instance = None
        self._zoning_instance = None
        self._demands_instance = None
        self._project_instance = None
        self._io_output_formatter_instance = None
        self._memory_output_formatter_instance = None
        self._initial_cost_instance = None
        self._activate_planitio_output_formatter = False
        self._activate_memory_output_formatter = False
        
        self._debug_info = debug_info
        
        if not standalone:
            raise Exception('Standalone argument can only be true at this time, server mode not yet supported')  
        self.__start_java__()
        self.__initialize_project__(project_path)
       
    def __start_java__(self):            
        """Start the gateway to Java 
        """  
        
        #find the location of this file, so that other directories can be located relative to it    
        dir_path = os.path.dirname(os.path.realpath(__file__))
        # Bootstrap the java gateway server
        if not GatewayState.gateway_is_running:
            # register dependencies (both for:
            #    - the IDE run, 
            #    - the RELEASE environment
            #    - a venv (virtual python) environment (which contains a bug that requires us to use a different path)
            dependencySet = {
                os.path.join(dir_path, GatewayConfig.IDE_SHARE_PATH,GatewayConfig.PLANIT_SHARE),
                os.path.join(dir_path, GatewayConfig.IDE_SHARE_PATH, GatewayConfig.PY4J_SHARE),
                os.path.join(GatewayConfig.RELEASE_SHARE_PATH, GatewayConfig.PLANIT_SHARE),
                os.path.join(GatewayConfig.VENV_RELEASE_SHARE_PATH,GatewayConfig.PLANIT_SHARE),                
                os.path.join(GatewayConfig.RELEASE_SHARE_PATH,GatewayConfig.PY4J_SHARE),
                os.path.join(GatewayConfig.VENV_RELEASE_SHARE_PATH, GatewayConfig.PY4J_SHARE)}
            dependencySeparator = ';'
            fullString = dependencySeparator.join(dependencySet)
            
            cmd = ['java', '-classpath', fullString, GatewayConfig.JAVA_GATEWAY_WRAPPER_CLASS]
            if self._debug_info: print('Java classpath: ' + fullString)
            GatewayState.planit_java_process = subprocess.Popen(cmd)           
             
            # now we  connect to the gateway
            GatewayState.python_2_java_gateway = JavaGateway()
            GatewayState.gateway_is_running = True            
                
            #TODO: Note we are not waiting for it to setup properly --> possibly considering some mechanism to wait for this to ensure proper connection!
            if self._debug_info: print('Java interface running with PID: '+ str(GatewayState.planit_java_process.pid))
        else:
            raise Exception('PLANit java interface already running, only a single instance allowed at this point')
    
    def __initialize_project__(self, project_path):
        """Initialize the project using the input file in the directory specified by project_path
        """
        if project_path == None:
            project_path = os.getcwd()
        self._project_instance = BaseWrapper(GatewayState.python_2_java_gateway.entry_point.initialiseSimpleProject(project_path))

        # The one macroscopic network, zoning, demand is created and populated and wrapped in a Python object
        # (Note1: to access public members in Java, we must collect it via the field method in the wrapper)
        # (Note2: since we only have a single network, demand, zoning, we do not have a wrapper for the fields, so we must access the methods directly
        self._network_instance = MacroscopicNetworkWrapper(self._project_instance.field("physicalNetworks").getFirstNetwork())
        # the one zoning is created and populated
        self._zoning_instance = ZoningWrapper(self._project_instance.field("zonings").getFirstZoning())
        # the one demands is created and populated
        self._demands_instance = DemandsWrapper(self._project_instance.field("demands").getFirstDemands())      
        self._initial_cost_instance = InitialCost()
        
        #PLANIT_IO output formatter is activated by default, MemoryOutputFormatter is off by default
        self._io_output_formatter_instance = PlanItOutputFormatterWrapper(self._project_instance.getDefaultOutputFormatter())
        self._activate_planitio_output_formatter = True
        
    def __del__(self):
        """Destructor of PLANit object which shuts down the connection to Java
        """
        self.__stop_java__()
        
    def __stop_java__(self):        
        """Cleans up the gateway in Java in case this has not been done yet. It assumes a single instance available in Python tied
        to a particular self. Only that instance is allowed to terminate the gateway.
        """          
        # Let the instance that instantiated the connection also terminate it automatically
        if GatewayState.gateway_is_running:
            # Check if the process has really terminated & force kill if not.           
            try:
                GatewayState.python_2_java_gateway.shutdown()
                GatewayState.planit_java_process.terminate()
                if (GatewayState.planit_java_process.poll() != None):
                    os.kill(GatewayState.planit_java_process.pid, 0)
                    GatewayState.planit_java_process.kill()
                # Wait for zombie process to provide post-mortem information. 
                # Only after this call the subprocess will be gone and we will not receive warnings
                # that it is still alive regardless of the fact we killed it zoom
                GatewayState.planit_java_process.wait()
                GatewayState.gateway_is_running = False
                GatewayState.python_2_java_gateway = None
                if self._debug_info: print ("Forced kill of PLANitJava interface")
            except OSError:
                if self._debug_info: print ("Terminated PLANitJava interface")   
            except:
                traceback.print_exc()         
        
    def __register_initial_costs__(self):   
        """Register the initial costs on the assignment
        """
        time_periods_external_id_set = self._initial_cost_instance.get_time_periods_external_id_set()
        
        if self._initial_cost_instance.get_default_initial_cost_file_location() != None:
            default_initial_cost_counterpart = self._project_instance.create_and_register_initial_link_segment_cost(self._network_instance.java, self._initial_cost_instance.get_default_initial_cost_file_location())
            default_initial_cost_wrapper = InitialCostWrapper(default_initial_cost_counterpart)
            self._assignment_instance.register_initial_link_segment_cost(default_initial_cost_wrapper.java)
            
        if len(time_periods_external_id_set) > 0:            
            time_period_counterparts = self._demands_instance.field("timePeriods").asSortedSetByStartTime()
            for time_period_counterpart in time_period_counterparts:
                time_period = TimePeriodWrapper(time_period_counterpart)
                time_period_external_id = time_period.get_external_id()
                if (time_period_external_id in time_periods_external_id_set):
                    initial_cost_file_location = self._initial_cost_instance.get_initial_cost_file_location_by_time_period_external_id(time_period_external_id)
                    initial_cost_counterpart = self._project_instance.create_and_register_initial_link_segment_cost(self._network_instance.java, initial_cost_file_location, time_period_counterpart)
                    initial_cost_wrapper = InitialCostWrapper(initial_cost_counterpart)
                    self._assignment_instance.register_initial_link_segment_cost(time_period.java, initial_cost_wrapper.java)       
                    
    def __getattr__(self, name):
        """ all methods invoked on the PLANit Java gateway wrapper as passed on to it without the user seeing the actual gateway. This is to be
        replaced by a more intricate interface which exposes only the properties users are allowed to configure to create a PLANit instance
        """        
        def method(*args): #collects the arguments of the function 'name' (wrapper function within getattr)                
            if GatewayState.gateway_is_running:
                java_name = GatewayUtils.to_camelcase(name)
                # pass all calls on to the underlying PLANit project java class which is obtained via the entry_point.getProject call
                return getattr(GatewayState.planit_project, java_name)(*args) # invoke without arguments
            else:
                raise Exception('PLANit java interface not available')      
        return method
        
    def set(self, assignment_component):
        """Set the traffic assignment component
        :param assignment_component the  assignment component
        """
        if isinstance(assignment_component, TrafficAssignment):
            assignment_counterpart = self._project_instance.create_and_register_traffic_assignment(assignment_component.value)
            self._assignment_instance = AssignmentWrapper(assignment_counterpart, self._network_instance)
            
    def activate(self, formatter_component):
        """Activate an output formatter
        :param formatter_component the formatter being set up
        """
        if formatter_component == OutputFormatter.PLANIT_IO:
            if (not self._activate_planitio_output_formatter):
                self._activate_planitio_output_formatter = True
                io_output_formatter_counterpart = self._project_instance.create_and_register_output_formatter(formatter_component.value)
                io_output_formatter = PlanItOutputFormatterWrapper(io_output_formatter_counterpart)
                self._io_output_formatter_instance = io_output_formatter
                        
        elif formatter_component == OutputFormatter.MEMORY:
            if (not self._activate_memory_output_formatter):
                self._activate_memory_output_formatter = True
                memory_output_formatter_counterpart =  self._project_instance.create_and_register_output_formatter(formatter_component.value)
                memory_output_formatter = MemoryOutputFormatterWrapper(memory_output_formatter_counterpart, self._demands_instance, self._network_instance)
                self._memory_output_formatter_instance = memory_output_formatter            
            
    def deactivate(self, formatter_component):
        """Deactivate an output formatter which has previously been activated
        :param formatter_component the formatter which has previously been activated
        """
        if formatter_component == OutputFormatter.PLANIT_IO:
            self._activate_planitio_output_formatter = False
            self._io_output_formatter_instance = None
        elif formatter_component == OutputFormatter.MEMORY:
            self._activate_memory_output_formatter = False
            self._memory_output_formatter_instance = None
        
    def run(self):
        """Run the traffic assignment.  Register any output formatters which have been set up
        """
        if (self._assignment_instance == None):
            raise Exception("Called plan_it.run() with no Traffic Assignment set")
        if (self._activate_planitio_output_formatter):
            self._assignment_instance.register_output_formatter(self._io_output_formatter_instance.java); 
        if (self._activate_memory_output_formatter):      
            self._assignment_instance.register_output_formatter(self._memory_output_formatter_instance.java)
        self.__register_initial_costs__()
        self._project_instance.execute_all_traffic_assignments()   
                                        
    @property
    def assignment(self):
        """ access to the assignment builder 
        """
        return self._assignment_instance  
    
    @property
    def network(self):
        """access to the network
        """
        return self._network_instance
    
    @property
    def demands(self):
        """access to the demands
        """
        return self._demands_instance
    
    @property
    def output(self):
        """access to PLANitIO output formatter
        """
        return self._io_output_formatter_instance
    
    @property
    def memory(self):
        """access to memory output formatter
        """
        return self._memory_output_formatter_instance
    
    @property
    def initial_cost(self):
        """access to initial cost
        """
        return self._initial_cost_instance


import os

from py4j.java_gateway import get_field
from planit import GatewayUtils
from planit import GatewayState
from planit import OutputType
from planit import RouteIdType
from planit import OutputProperty
from planit import PhysicalCost
from planit import VirtualCost
from planit import Smoothing

class BaseWrapper(object):
    """ Base wrapper class which always holds a java counter part instance and a generic way to pass on method calls to the encapsulated 
        Java instance. This allows one to use the Python naming conventions, i.e. get_x instead of getX() this is automatically catered for
        in these wrappers.
    """

    def __init__(self, java_counterpart):
        """Top-level constructor which defines the Java counterpart to each wrapper
            :param   the Java counterpart to the current object
        """
        self._java_counterpart = java_counterpart

    def __getattr__(self, name):
        """All methods invoked on the assignment wrapper are passed on to the Java equivalent class after transforming the method to
        Java style coding convention
        """            
        def method(*args): #collects the arguments of the function 'name' (wrapper function within getattr)    
            java_name = GatewayUtils.to_camelcase(name)
            # pass all calls on to the underlying PLANit project java class which is obtained via the entry_point.getProject call
            if( args ):
                return getattr(self._java_counterpart, java_name)(*args)
            else:
                return getattr(self._java_counterpart, java_name)()               
        return method
    
    @property
    def java(self):
        """ access to the underlying Java object if required
        """
        if (self._java_counterpart == None):
            raise Exception("No Java counterpart has been found for " + self.__class__.__name__)
        return self._java_counterpart 
    
    def field(self, fieldName):
        """ collect a publicly available member on the java object
        """
        return get_field(self.java, fieldName)

class AssignmentWrapper(BaseWrapper):
    """ Wrapper around the Java traffic assignment builder class instance
    """
    
    def __init__(self, java_counterpart, network_instance):
        super().__init__(java_counterpart)
        self._output_configuration = OutputConfigurationWrapper(self.get_output_configuration()) # collect the output configuration from Java
        self._gap_function = GapFunctionWrapper(self.get_gap_function()) # collect the gap function from Java
        self._network_instance = network_instance
        self._physical_cost_instance = BPRCostWrapper(self.get_physical_cost(), self._network_instance) 
        self._virtual_cost_instance = VirtualCostWrapper(self.get_virtual_cost()) 
        self._smoothing = SmoothingWrapper(self.get_smoothing())
        
        # initialize in case they have defaults available
        #=======================================================================
        # self._link_output_type_configuration = LinkOutputTypeConfigurationWrapper(self._output_configuration.get_output_type_configuration(self.__create_java_output_type(OutputType.LINK)))
        # self._origin_destination_output_type_configuration = OriginDestinationOutputTypeConfigurationWrapper(self._output_configuration.get_output_type_configuration(self.__create_java_output_type(OutputType.OD)))   
        # self._path_output_type_configuration = PathOutputTypeConfigurationWrapper(self._output_configuration.get_output_type_configuration(self.__create_java_output_type(OutputType.PATH)))
        #=======================================================================
        
        for output_type in OutputType:
            output_type_instance = self.__create_java_output_type(output_type)
            if self.is_output_type_active(output_type_instance):
                if output_type.value == "LINK":
                    self._link_output_type_configuration = LinkOutputTypeConfigurationWrapper(self._java_counterpart.activateOutput(output_type_instance))
                elif output_type.value == "OD":
                    self._origin_destination_output_type_configuration = OriginDestinationOutputTypeConfigurationWrapper(self._java_counterpart.activateOutput(output_type_instance))
                elif output_type.value == 'PATH':
                    self._path_output_type_configuration = PathOutputTypeConfigurationWrapper(self._java_counterpart.activateOutput(output_type_instance))
        
        
        
    def __create_java_output_type(self, output_type):
        """ create an output type enum suitable to pass to java 
        """   
        return GatewayState.python_2_java_gateway.entry_point.createEnum(output_type.java_class_name(), output_type.value)
    
    def set(self, assignment_component):
        """ Configure an assignment component on this assignment instance. Note that all these go via the traffic assignment builder in Java
            although we hide that on the Python side to not over-complicate things for the average user. We accept PhysicalCost, VirtualCost and 
            Smoothing choices at this point
        """    
        
        if isinstance(assignment_component, PhysicalCost):
            if (assignment_component == PhysicalCost.BPR):
                self._physical_cost_instance = BPRCostWrapper(self.create_and_register_physical_cost(assignment_component.value), self._network_instance) 
            else:
                raise Exception('Unrecognized link cost function ' + assignment_component.type + ' cannot be set on assignment instance')
        elif isinstance(assignment_component, VirtualCost):
            self._virtual_cost_instance = VirtualCostWrapper(self.create_and_register_virtual_cost(assignment_component.value))
        elif isinstance(assignment_component, Smoothing):
            self._smoothing_instance = SmoothingWrapper(self.create_and_register_smoothing(assignment_component.value))
        else:
            raise Exception('Unrecognized component ' + assignment_component.type + ' cannot be set on assignment instance')
         
    def activate_output(self, output_type : OutputType):
        """ Activate different output types on the assignment. 
            Pass on to Java not as an Enum as Py4J does not seem to properly handle this at this stage
            instead we pass on the enum string which on the Java side is converted into the proper enum instead           
            :param output_type Python enum of available output types
        """ 
        # collect an enum instance by collecting the <package>.<class_name> string from the Output type enum
        output_type_instance = self.__create_java_output_type(output_type)
        if not self.is_output_type_active(output_type_instance):
            if output_type.value == "LINK":
                self._link_output_type_configuration = LinkOutputTypeConfigurationWrapper(self._java_counterpart.activateOutput(output_type_instance))
            elif output_type.value == "OD":
                self._origin_destination_output_type_configuration = OriginDestinationOutputTypeConfigurationWrapper(self._java_counterpart.activateOutput(output_type_instance))
            elif output_type.value == 'PATH':
                self._path_output_type_configuration = PathOutputTypeConfigurationWrapper(self._java_counterpart.activateOutput(output_type_instance))
            else:
                raise ValueError("Attempted to deactivate unknown output type " + output_type.value)     
    
    def deactivate_output(self, output_type: OutputType):   
        """Deactivate specified output type on the assignment
        :param output_type Output type to be deactivated
        """ 
        output_type_instance = self.__create_java_output_type(output_type)
        if self.is_output_type_active(output_type_instance):
            if output_type.value == "LINK":
                self._link_output_type_configuration = LinkOutputTypeConfigurationWrapper(self._java_counterpart.deactivateOutput(output_type_instance))
            elif output_type.value == "OD":
                self._origin_destination_output_type_configuration = OriginDestinationOutputTypeConfigurationWrapper(self._java_counterpart.deactivateOutput(output_type_instance))
            elif output_type.value == 'PATH':
                self._path_output_type_configuration = PathOutputTypeConfigurationWrapper(self._java_counterpart.deactivateOutput(output_type_instance))
            else:
                raise ValueError("Attempted to activate unknown output type " + output_type.value)     
  
    @property
    def gap_function(self):
        """Access to current gap function
        """
        return self._gap_function
    
    @property
    def link_configuration(self):
        """Access to current link output type configuration, if not activated, it is activated so it can be configured
        """
        self.activate_output(OutputType.LINK)
        return self._link_output_type_configuration
    
    @property
    def od_configuration(self):
        """Access to current origin-destination output type configuration, if not activated, it is activated so it can be configured
        """
        self.activate_output(OutputType.OD)
        return self._origin_destination_output_type_configuration
    
    @property
    def output_configuration(self):
        """Access to current output configuration
        """
        return self._output_configuration    
    
    @property
    def path_configuration(self):
        """Access to current path output type configuration, if not activated, it is activated so it can be configured
        """
        self.activate_output(OutputType.PATH)
        return self._path_output_type_configuration
           
    @property
    def physical_cost(self):
        """Access to the physical cost wrapper
        """
        return self._physical_cost_instance
    
    @property
    def smoothing(self):
        """Access to the smoothing wrapper
        """
        return self._smoothing_instance    
    
    @property
    def virtual_cost(self):
        """Access to the virtual cost wrapper
        """
        return self._virtual_cost_instance
    
    
class DemandsWrapper(BaseWrapper):
    """ Wrapper around the Java Demands class instance
    """
    
    def __init__(self, java_counterpart):
        super().__init__(java_counterpart)    
        
class GapFunctionWrapper(BaseWrapper):
    """ Wrapper around the Java GapFunction class instance
    """
    
    def __init__(self, java_counterpart):
        super().__init__(java_counterpart)
        self._stop_criterion = StopCriterionWrapper(self.get_stop_criterion())  # collect the stop criterion from Java
        
    @property
    def stop_criterion(self):
        return self._stop_criterion

class InitialCostWrapper(BaseWrapper):
    """ Wrapper around the Java initial cost class instance
    """
    
    def __init__(self, java_counterpart):
        super().__init__(java_counterpart)
        
class LinkSegmentWrapper(BaseWrapper):
    """ Wrapper around the Java LinkSegments class
    """
    
    def __init__(self, java_counterpart):
        super().__init__(java_counterpart)
        
class LinkSegmentsWrapper(BaseWrapper):
    """ Wrapper around the Java LinkSegments class
    """
    
    def __init__(self, java_counterpart):
        super().__init__(java_counterpart)
        
class LinkSegmentExpectedResultsDtoWrapper(BaseWrapper):
    """ Wrapper around the Java Link Segment Expected Results DTO class instance
    """
    
    def __init__(self, java_counterpart):
        super().__init__(java_counterpart)

class MacroscopicLinkSegmentWrapper(BaseWrapper):
    """ Wrapper around the MacroscopicLinkSegmentType class instance
    """
     
    def __init__(self, java_counterpart):
        super().__init__(java_counterpart)      
 
class MacroscopicLinkSegmentTypeWrapper(BaseWrapper):
    """ Wrapper around the MacroscopicLinkSegmentType class instance
    """
     
    def __init__(self, java_counterpart):
        super().__init__(java_counterpart)      
 
class MacroscopicNetworkWrapper(BaseWrapper):
    """ Wrapper around the Java physical network class instance
    """
    
    def __init__(self, java_counterpart):
        super().__init__(java_counterpart)
        
class MemoryOutputIteratorWrapper(BaseWrapper):
    """Wrapper class around MemoryOutputIterator class
    """
    
    def __init__(self, java_counterpart):
        super().__init__(java_counterpart)
       
class ModeWrapper(BaseWrapper):
    """ Wrapper around the Java mode class instance
    """
    
    def __init__(self, java_counterpart):
        super().__init__(java_counterpart)

class ModesWrapper(BaseWrapper):
    """ Wrapper around the Java modes class instance
    """
    
    def __init__(self, java_counterpart):
        super().__init__(java_counterpart)

class OutputConfigurationWrapper(BaseWrapper):
    """ Wrapper around the Java output configuration class instance
    """
    
    def __init__(self, java_counterpart):
        super().__init__(java_counterpart)
 
class OutputFormatterWrapper(BaseWrapper):
    """ Wrapper around the Java OutputFormatter wrapper class instance
    """
    
    def __init__(self, java_counterpart):
        super().__init__(java_counterpart)

class OutputTypeConfigurationWrapper(BaseWrapper): 
    """ Wrapper around the Java link output type configuration class instance
    """
     
    def __init__(self, java_counterpart):
        super().__init__(java_counterpart)
        
    def add(self, output_property : OutputProperty):
        """Add an output type property to the current output type configuration
        """
        output_property_instance = GatewayState.python_2_java_gateway.entry_point.createEnum(output_property.java_class_name(), output_property.value)
        self._java_counterpart.addProperty(output_property_instance)
        
    def remove(self, output_property : OutputProperty):
        """Remove an output type property from the current output type configuration
        """
        output_property_instance = GatewayState.python_2_java_gateway.entry_point.createEnum(output_property.java_class_name(), output_property.value)
        return self._java_counterpart.removeProperty(output_property_instance)
        
    def remove_all_properties(self):
        self._java_counterpart.removeAllProperties()
        
class PhysicalCostWrapper(BaseWrapper):
    """ Wrapper around the Java physical cost class instance
    """
    
    def __init__(self, java_counterpart):
        super().__init__(java_counterpart)
  
class PlanItInputBuilderWrapper(BaseWrapper):
    """ Wrapper around the Java InputBuilderListener class instance
    """
    
    def __init__(self, java_counterpart):
        super().__init__(java_counterpart)    
  
class SmoothingWrapper(BaseWrapper):
    """ Wrapper around the Java Smoothing class instance
    """
    
    def __init__(self, java_counterpart):
        super().__init__(java_counterpart)
        
class StopCriterionWrapper(BaseWrapper):
    """ Wrapper around the Java StopCriterion class instance
    """
    
    def __init__(self, java_counterpart):
        super().__init__(java_counterpart)
    
class TimePeriodWrapper(BaseWrapper):
    """ Wrapper around the Java time period class instance
    """
    
    def __init__(self, java_counterpart):
        super().__init__(java_counterpart)

class TimePeriodsWrapper(BaseWrapper):
    """ Wrapper around the Java time periods class instance
    """
    
    def __init__(self, java_counterpart):
        super().__init__(java_counterpart)

class VirtualCostWrapper(BaseWrapper):
    """ Wrapper around the Java assignment class instance
    """    
    
    def __init__(self, java_counterpart):
        super().__init__(java_counterpart)

class ZoningWrapper(BaseWrapper):
    """ Wrapper around the Java Zoning class instance
    """    
    
    def __init__(self, java_counterpart):
        super().__init__(java_counterpart)
        
class BPRCostWrapper(PhysicalCostWrapper):
    """Wrapper around the BPRLinkTravelTimeCost instance
    """
    
    def __init__(self, java_counterpart, network_instance):
        super().__init__(java_counterpart)
        self._network_instance = network_instance
        modes_counterpart = self._network_instance.field("modes")
        self._modes_instance = ModesWrapper(modes_counterpart)
        link_segments_counterpart = self._network_instance.field("linkSegments")
        self._link_segments_instance = LinkSegmentsWrapper(link_segments_counterpart)
        
    def set_default_parameters(self, alpha, beta, mode_external_id=None, link_segment_type_external_id=None):
        """Set the default BPR functions parameters 
        :param alpha value of alpha parameter
        :param beta value of beta parameter
        :param mode_external_id if included, default parameters only apply to this mode
        :param link_segment_type_external_id if included, default parameters apply to this link segment type and mode
        """
        if (mode_external_id == None):
            self._java_counterpart.setDefaultParameters(alpha, beta)
        else:
            mode_counterpart = self._modes_instance.get_mode_by_external_id(mode_external_id, True)
            if (link_segment_type_external_id == None):
                self._java_counterpart.setDefaultParameters(mode_counterpart, alpha, beta)
            else:
                link_segment_type_counterpart = self._network_instance.get_macroscopic_link_segment_type_by_external_id(link_segment_type_external_id, True)
                link_segment_type_instance = MacroscopicLinkSegmentTypeWrapper(link_segment_type_counterpart)
                self.setDefaultParameters(link_segment_type_instance.java, mode_counterpart, alpha, beta)
                
    def set_parameters(self, alpha, beta, mode_external_id, link_segment_external_id):
        """Set the default BPR functions parameters 
        :param alpha value of alpha parameter
        :param beta value of beta parameter
        :param mode_external_id, parameters only apply to this mode
        :param link_segment_external_id, parameters apply to this link segment 
        """        
        link_segment_counterpart = self._link_segments_instance.get_link_segment_by_external_id(link_segment_external_id, True)
        link_segment_instance = LinkSegmentWrapper(link_segment_counterpart)
        mode_counterpart = self._modes_instance.get_mode_by_external_id(mode_external_id, True)
        self._java_counterpart.setParameters(link_segment_instance.java, mode_counterpart, alpha, beta)

class MemoryOutputFormatterWrapper(OutputFormatterWrapper):
    """ Wrapper around the Java PlanItOutputFormatter class instance
    """
    
    def __init__(self, java_counterpart, demands_instance, network_instance):
        """
        :param self this object
        :param java_counterpart Java counterpart for MemoryOutputFormatter object
        :param project_instance the instance of the project being run
        """
        super().__init__(java_counterpart)
        self._demands_instance = demands_instance
        self._network_instance = network_instance
                   
    def iterator(self, mode_external_id, time_period_external_id, no_iterations, output_type):
        """Return the  wrapper for MemoryOutputIterator object for this MemoryOutputFormatter
        :param mode_external_id the external Id of the current mode
        :param time_period_external_id the external Id of the current time period
        :param no_iterations the iteration the output iterator applies to
        :param output_type the output type for the current output
        :return the wrapper for the memory output iterator
        """
        time_periods_counterpart = self._demands_instance.field("timePeriods")
        time_periods = TimePeriodsWrapper(time_periods_counterpart)
        time_period_counterpart = time_periods.get_time_period_by_external_id(time_period_external_id, True);
        time_period = TimePeriodWrapper(time_period_counterpart)        
        modes_counterpart = self._network_instance.field("modes")
        modes = ModesWrapper(modes_counterpart)
        mode_counterpart = modes.get_mode_by_external_id(mode_external_id, True)
        mode = ModeWrapper(mode_counterpart)       
        output_type_instance = GatewayState.python_2_java_gateway.entry_point.createEnum(output_type.java_class_name(), output_type.value)
        memory_output_iterator_counterpart = self._java_counterpart.getIterator(mode.java, time_period.java, no_iterations, output_type_instance)
        memory_output_iterator = MemoryOutputIteratorWrapper(memory_output_iterator_counterpart)
        return memory_output_iterator
   
    def get_position_of_output_value_property(self, output_type, output_property):
        """Returns the position in the results array of the specified output value property
        :param output_type the output type for the current output
        :param output_property the specified output value property
        :result the position in the results array of the specified property
        """
        output_type_instance = GatewayState.python_2_java_gateway.entry_point.createEnum(output_type.java_class_name(), output_type.value)
        output_property_instance = GatewayState.python_2_java_gateway.entry_point.createEnum(output_property.java_class_name(), output_property.value)
        position = self._java_counterpart.getPositionOfOutputValueProperty(output_type_instance, output_property_instance)
        return position
    
    def get_position_of_output_key_property(self, output_type, output_property):
        """Returns the position in the results array of the specified output key property
          :param output_type the output type for the current output
          :param output_property the specified output property
          :result the position in the results array of the specified key property
        """
        output_type_instance = GatewayState.python_2_java_gateway.entry_point.createEnum(output_type.java_class_name(), output_type.value)
        output_property_instance = GatewayState.python_2_java_gateway.entry_point.createEnum(output_property.java_class_name(), output_property.value)
        position = self._java_counterpart.getPositionOfOutputKeyProperty(output_type_instance, output_property_instance)
        return position
        
class PlanItOutputFormatterWrapper(OutputFormatterWrapper):
    """ Wrapper around the Java PlanItOutputFormatter class instance
    """
    
    def __init__(self, java_counterpart):
        super().__init__(java_counterpart)
        # Initialize the output project path to the current run directory
        # Modellers may overwrite this default later
        project_path = os.getcwd()
        self.set_output_directory(project_path)        
              
class LinkOutputTypeConfigurationWrapper(OutputTypeConfigurationWrapper):
    """ Wrapper around the Java link output type configuration class instance
    """
     
    def __init__(self, java_counterpart):
        super().__init__(java_counterpart)
       
class OriginDestinationOutputTypeConfigurationWrapper(OutputTypeConfigurationWrapper):
    """ Wrapper around the Java origin-destination output type configuration class instance
    """
     
    def __init__(self, java_counterpart):
        super().__init__(java_counterpart)

    def activate(self, od_skim_sub_output_type):
        """Activate an OD skim output type
        """
        od_skim_sub_output_type_instance = GatewayState.python_2_java_gateway.entry_point.createEnum(od_skim_sub_output_type.java_class_name(), od_skim_sub_output_type.value)
        self._java_counterpart.activateOdSkimOutputType(od_skim_sub_output_type_instance)
 
    def deactivate(self, od_skim_sub_output_type):
        """Deactivate an OD skim output type
        """
        od_skim_sub_output_type_instance = GatewayState.python_2_java_gateway.entry_point.createEnum(od_skim_sub_output_type.java_class_name(), od_skim_sub_output_type.value)
        self._java_counterpart.deactivateOdSkimOutputType(od_skim_sub_output_type_instance)         
        
class PathOutputTypeConfigurationWrapper(OutputTypeConfigurationWrapper):
    """ Wrapper around the Java path output type configuration class instance
    """
     
    def __init__(self, java_counterpart):
        super().__init__(java_counterpart)
        
    def set_path_id_type(self,  route_id_type : RouteIdType):
        route_id_type_instance = GatewayState.python_2_java_gateway.entry_point.createEnum(route_id_type.java_class_name(), route_id_type.value)
        self._java_counterpart.setPathIdType(route_id_type_instance)


# version that matches the PLANit-Java version
# Used to:
# 1) Construct the correct name of the used PLANit-jar files
# 2) Construct the version of the Python module via setup.py (hence use of globals)
__planit_version__ = "0.1.0"
__py4j_version__ = "0.10.9"

class Version:
    """ access to global versions in OO style
    """
    planit = __planit_version__
    py4j = __py4j_version__


"""This class exists to take the logic for setting up the initial costs out of the PLANit class.  It does not wrap any Java object.  
This class is instantiated as a member of the PLANit object.  It allows top level calls to have the signature "plan_it.initial_cost.set(..."
"""

class InitialCost:
    
    def __init__(self):
        """Initializer for the InitialCosts class
        """
        self._default_initial_cost_file_location = None
        self._initial_cost_location_dictionary = {}
        
    def set(self, initial_cost_file_location, time_period_external_id=None):
        """Set an initial cost file location
        :param initial_cost_file_location location of an initial cost file
        :param time_period_external_id external id of the time period for which these initial costs apply
        """
        if (time_period_external_id == None):
            self._default_initial_cost_file_location = initial_cost_file_location
        else:
            self._initial_cost_location_dictionary[time_period_external_id] = initial_cost_file_location
            
    def get_initial_cost_file_location_by_time_period_external_id(self, time_period_external_id):
        return self._initial_cost_location_dictionary[time_period_external_id]
    
    def get_time_periods_external_id_set(self):
        return self._initial_cost_location_dictionary.keys()
    
    def get_default_initial_cost_file_location(self):
        return self._default_initial_cost_file_location


import re
import os
import sys

from planit import Version

class GatewayConfig(object):   

    # Currently we provide the paths for the local environment AND production environment
    # TODO there should be a mechanism such that we only need one type of definition
    
    RELEASE_SHARE_PATH = os.path.join(os.path.split(sys.executable)[0], 'share')
    IDE_SHARE_PATH = os.path.join('..', '..', 'share')
    # VENV (virtual environment) for Python erroneously results in sys.executable pointing NOT to the root
    # directory of the virtual environment, but to ./Scripts. Hence, we must account for that by alos looking
    # one directory upward.
    # This (as far as I know) is a bug in venv!
    VENV_RELEASE_SHARE_PATH = os.path.join(os.path.split(sys.executable)[0], '..', 'share')
    
    PLANIT_SHARE = os.path.join('planit', '*')
    PY4J_SHARE = os.path.join('py4j', '*')
    
    # the main entry point of the Java gateway implementation for PLANit
    JAVA_GATEWAY_WRAPPER_CLASS =  'org.planit.python.PLANitJ2Py'
    
    
class GatewayState(object):
    #Create a static variable which flags if the java server already is running or not
    gateway_is_running = False
    planit_java_process = None
    # The actual gateway to pass on requests to once the gateway server is known to be running
    python_2_java_gateway = None
    # will contain reference to the Java project instance once the gateway is up and running        
    planit_project = None 


class GatewayUtils(object):
 
    @staticmethod
    def to_camelcase(s):
            """ convert a Python style string into a Java style string regarding method calls and variable names. Especially useful to avoid
            having to call Java functions as Java functions but one can call them as Python functions which are dynamically changed to their
            Java counterparts
            """
            return re.sub(r'(?!^)_([a-zA-Z])', lambda m: m.group(1).upper(), s) 




from setuptools import setup                                                # setuptools is our method for building the module
from setuptools import find_packages

from glob import glob

import os

# VERSION
# parse the __planit_version__ variable from python script
VERSION_PATH = os.path.join("src", "planit", "version.py")
exec(compile(open(VERSION_PATH).read(),VERSION_PATH, "exec"))
VERSION = __planit_version__
LICENSE_LOCATION = "https://trafficplanit.github.io/PLANitManual/docs/licenses/"

# RESOURCES
# pars all jars from resource dir
RESOURCE_DIR = "share/planit"
RESOURCE_JAR_FILE_NAMES = glob(RESOURCE_DIR+'/**')                          

setup(
    name="PLANit-Python",
    version= VERSION,
    description="Python API for traffic assignment using PLANit",
    long_description="PLANit-Python enables Python programs running in "
                     "a Python interpreter to configure and run "
                     "a PLANit traffic assignment. It uses Py4J to "
                     "access the underlying Java API that can be used "
                     "for the same purpose",                     
    url="https://trafficplanit.github.io/PLANitManual",
    author="Mark Raadsen",
    author_email="mark.raadsen@sydney.edu.au",                              
    license="modified BSD License (see "+LICENSE_LOCATION+")",              # adopted license
    classifiers=[                                                           # meta information regarding this module
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Java",
        "Topic :: Scientific/Engineering",        
    ],
    packages=["planit", "test_utils"],                                      # we only include these packages  
    package_dir={"": "src"},                                                # indicate the ./src directory is where to find packages rather than this ""
    data_files=[(RESOURCE_DIR, RESOURCE_JAR_FILE_NAMES)],                          # copy all jars in rsc dir as data_files in module
    install_requires=[
          'py4j>='+__py4j_version__,                                        # python installation for py4j
      ],    

)


"""Package initialization file for the planit package
"""
from .enums import Network
from .enums import OutputFormatter
from .enums import OutputProperty
from .enums import OutputType
from .enums import PhysicalCost
from .enums import RouteIdType
from .enums import Smoothing
from .enums import TrafficAssignment
from .enums import VirtualCost
from .enums import ODSkimSubOutputType
from .version import Version
from .gateway import GatewayConfig
from .gateway import GatewayState
from .gateway import GatewayUtils
from .wrappers import BaseWrapper
from .wrappers import AssignmentWrapper
from .wrappers import DemandsWrapper
from .wrappers import GapFunctionWrapper
from .wrappers import ModeWrapper
from .wrappers import OutputConfigurationWrapper
from .wrappers import OutputFormatterWrapper
from .wrappers import PlanItOutputFormatterWrapper
from .wrappers import MemoryOutputFormatterWrapper
from .wrappers import OutputTypeConfigurationWrapper
from .wrappers import LinkOutputTypeConfigurationWrapper
from .wrappers import OriginDestinationOutputTypeConfigurationWrapper
from .wrappers import PathOutputTypeConfigurationWrapper
from .wrappers import PhysicalCostWrapper
from .wrappers import InitialCostWrapper
from .wrappers import MacroscopicNetworkWrapper
from .wrappers import PlanItInputBuilderWrapper
from .wrappers import SmoothingWrapper
from .wrappers import StopCriterionWrapper
from .wrappers import TimePeriodWrapper
from .wrappers import VirtualCostWrapper
from .wrappers import ZoningWrapper
from .wrappers import TimePeriodsWrapper
from .wrappers import ModesWrapper
from .wrappers import LinkSegmentExpectedResultsDtoWrapper
from .wrappers import MemoryOutputIteratorWrapper
from .initial_cost import InitialCost
from .PLANit import PLANit